@@ -27,12 +27,11 @@ Approach:
 3. Verify checksums.
 """
 
-from urllib.error import (URLError, HTTPError)
 from urllib.request import urlopen
 import codecs
 import functools
 import json
-import os.path
+import os
 import re
 import subprocess
 import sys
@@ -45,25 +44,31 @@ import debian.deb822
 
 from contextlib import closing
 
+from pathlib import Path
+
 from ubuntutools.config import UDTConfig
-from ubuntutools.lp.lpapicache import (Launchpad, Distribution, PersonTeam, Project,
-                                       SourcePackagePublishingHistory)
+from ubuntutools.lp.lpapicache import (Launchpad,
+                                       Distribution,
+                                       PersonTeam,
+                                       Project,
+                                       SourcePackagePublishingHistory,
+                                       HTTPError)
 from ubuntutools.lp.udtexceptions import (PackageNotFoundException,
                                           SeriesNotFoundException,
                                           PocketDoesNotExistError,
                                           InvalidDistroValueError)
-from ubuntutools.misc import (download, download_text, verify_file_checksum, verify_file_checksums)
+from ubuntutools.misc import (download,
+                              download_bytes,
+                              verify_file_checksum,
+                              verify_file_checksums,
+                              DownloadError,
+                              NotFoundError)
 from ubuntutools.version import Version
 
 import logging
 Logger = logging.getLogger(__name__)
 
 
-class DownloadError(Exception):
-    "Unable to pull a source package"
-    pass
-
-
 class Dsc(debian.deb822.Dsc):
     "Extend deb822's Dsc with checksum verification abilities"
 
@@ -83,11 +88,12 @@ class Dsc(debian.deb822.Dsc):
 
     def verify_file(self, pathname):
         "Verify that pathname matches the checksums in the dsc"
-        if not os.path.isfile(pathname):
+        p = Path(pathname)
+        if not p.is_file():
             return False
         alg, checksums = self.get_strongest_checksum()
-        size, digest = checksums[os.path.basename(pathname)]
-        return verify_file_checksum(pathname, alg, digest, size)
+        size, digest = checksums[p.name]
+        return verify_file_checksum(p, alg, digest, size)
 
     def compare_dsc(self, other):
         """Check whether any files in these two dscs that have the same name
@@ -139,7 +145,7 @@ class SourcePackage(ABC):
         """
         dscfile = kwargs.get('dscfile')
         mirrors = kwargs.get('mirrors', ())
-        workdir = kwargs.get('workdir', '.')
+        workdir = kwargs.get('workdir')
         series = kwargs.get('series')
         pocket = kwargs.get('pocket')
         status = kwargs.get('status')
@@ -158,11 +164,11 @@ class SourcePackage(ABC):
         self.source = package
         self.binary = None
         self.try_binary = try_binary
-        self.workdir = workdir
+        self.workdir = Path(workdir) if workdir else Path('.')
         self._series = series
         self._pocket = pocket
         self._status = status
-        self._dsc_source = dscfile
+        self._dsc_source = Path(dscfile) if dscfile else None
         self._verify_signature = verify_signature
 
         # Cached values:
@@ -180,7 +186,7 @@ class SourcePackage(ABC):
 
         # If provided a dscfile, process it now to set our source and version
         if self._dsc_source:
-            self._dsc = Dsc(download_text(self._dsc_source, mode='rb'))
+            self._dsc = Dsc(download_bytes(self._dsc_source))
             self.source = self._dsc['Source']
             self._version = Version(self._dsc['Version'])
             self._check_dsc_signature()
@@ -267,12 +273,12 @@ class SourcePackage(ABC):
     @property
     def dsc_name(self):
         "Return the source package dsc filename for the given package"
-        return '%s_%s.dsc' % (self.source, self.version.strip_epoch())
+        return f'{self.source}_{self.version.strip_epoch()}.dsc'
 
     @property
     def dsc_pathname(self):
         "Return the dsc_name, with the workdir path"
-        return os.path.join(self.workdir, self.dsc_name)
+        return str(self.workdir / self.dsc_name)
 
     @property
     def dsc(self):
@@ -282,10 +288,9 @@ class SourcePackage(ABC):
                 raise RuntimeError('Internal error: we have a dsc file but dsc not set')
             urls = self._source_urls(self.dsc_name)
             with tempfile.TemporaryDirectory() as d:
-                tmpdsc = os.path.join(d, self.dsc_name)
+                tmpdsc = Path(d) / self.dsc_name
                 self._download_file_from_urls(urls, tmpdsc)
-                with open(tmpdsc, 'rb') as f:
-                    self._dsc = Dsc(f.read())
+                self._dsc = Dsc(tmpdsc.read_bytes())
                 self._check_dsc_signature()
         return self._dsc
 
@@ -297,13 +302,13 @@ class SourcePackage(ABC):
     def getArchive(self):
         return self.getDistribution().getArchive()
 
-    def _mirror_url(self, mirror, filename):
+    def _mirror_url(self, mirror, component, filename):
         "Build a source package URL on a mirror"
         if self.source.startswith('lib'):
             group = self.source[:4]
         else:
             group = self.source[0]
-        return os.path.join(mirror, 'pool', self.component, group,
+        return os.path.join(mirror, 'pool', component, group,
                             self.source, filename)
 
     def _archive_servers(self):
@@ -318,16 +323,16 @@ class SourcePackage(ABC):
     def _source_urls(self, name):
         "Generator of sources for name"
         if self._dsc_source:
-            yield os.path.join(os.path.dirname(self._dsc_source), name)
+            yield str(self._dsc_source.parent / name)
         for server in self._archive_servers():
-            yield self._mirror_url(server, name)
+            yield self._mirror_url(server, self.component, name)
         if self.lp_spph.sourceFileUrl(name):
             yield self.lp_spph.sourceFileUrl(name)
 
     def _binary_urls(self, name, bpph):
         "Generator of URLs for name"
         for server in self._archive_servers():
-            yield self._mirror_url(server, name)
+            yield self._mirror_url(server, bpph.getComponent(), name)
         if bpph.binaryFileUrl(name):
             yield bpph.binaryFileUrl(name)
         if bpph.getUrl():
@@ -362,33 +367,35 @@ class SourcePackage(ABC):
             Logger.warning('Signature on %s could not be verified' % self.dsc_name)
 
     def _verify_file(self, pathname, dscverify=False, sha1sum=None, sha256sum=None, size=0):
-        if not os.path.exists(pathname):
+        p = Path(pathname)
+        if not p.exists():
             return False
-        if dscverify and not self.dsc.verify_file(pathname):
+        if dscverify and not self.dsc.verify_file(p):
             return False
         checksums = {}
         if sha1sum:
             checksums['SHA1'] = sha1sum
         if sha256sum:
             checksums['SHA256'] = sha256sum
-        if not verify_file_checksums(pathname, checksums, size):
+        if not verify_file_checksums(p, checksums, size):
             return False
         return True
 
     def _download_file(self, url, filename, size=0, dscverify=False, sha1sum=None, sha256sum=None):
         "Download url to filename; will be put in workdir unless filename is absolute path."
-        if os.path.isabs(filename):
-            pathname = filename
+        if Path(filename).is_absolute():
+            p = Path(filename).expanduser().resolve()
         else:
-            pathname = os.path.join(self.workdir, filename)
+            p = self.workdir / filename
 
-        if self._verify_file(pathname, dscverify, sha1sum, sha256sum, size):
-            Logger.info('Using existing file %s', filename)
+        can_verify = any((dscverify, sha1sum, sha256sum))
+        if can_verify and self._verify_file(p, dscverify, sha1sum, sha256sum, size):
+            Logger.info(f'Using existing file {p}')
             return True
 
-        download(url, pathname, size)
+        download(url, p, size)
 
-        return self._verify_file(pathname, dscverify, sha1sum, sha256sum, size)
+        return self._verify_file(p, dscverify, sha1sum, sha256sum, size)
 
     def _download_file_from_urls(self, urls, filename, size=0, dscverify=False,
                                  sha1sum=None, sha256sum=None):
@@ -398,15 +405,12 @@ class SourcePackage(ABC):
                 if self._download_file(url, filename, size, dscverify=dscverify,
                                        sha1sum=sha1sum, sha256sum=sha256sum):
                     return
-            except HTTPError as e:
-                if e.code == 404:
-                    # It's ok if the file isn't found, just try the next url
-                    Logger.debug("File not found at %s" % url)
-                else:
-                    Logger.error('HTTP Error %i: %s', e.code, str(e))
-            except URLError as e:
-                Logger.error('URL Error: %s', e.reason)
-        raise DownloadError('Failed to download %s' % filename)
+            except NotFoundError:
+                # It's ok if the file isn't found, just try the next url
+                Logger.debug(f'File not found at {url}')
+            except DownloadError as e:
+                Logger.error(f'Download Error: {e}')
+        raise DownloadError(f'Failed to download {filename}')
 
     def pull_dsc(self):
         '''DEPRECATED
@@ -424,8 +428,7 @@ class SourcePackage(ABC):
 
     def pull(self):
         "Pull into workdir"
-        with open(self.dsc_pathname, 'wb') as f:
-            f.write(self.dsc.raw_text)
+        Path(self.dsc_pathname).write_bytes(self.dsc.raw_text)
         for entry in self.dsc['Files']:
             name = entry['name']
             urls = self._source_urls(name)
@@ -467,8 +470,7 @@ class SourcePackage(ABC):
         """Verify that the source package in workdir matches the dsc.
         Return boolean
         """
-        return all(self.dsc.verify_file(os.path.join(self.workdir,
-                                                     entry['name']))
+        return all(self.dsc.verify_file(self.workdir / entry['name'])
                    for entry in self.dsc['Files'])
 
     def verify_orig(self):
@@ -476,8 +478,7 @@ class SourcePackage(ABC):
         Return boolean
         """
         orig_re = re.compile(r'.*\.orig(-[^.]+)?\.tar\.[^.]+$')
-        return all(self.dsc.verify_file(os.path.join(self.workdir,
-                                                     entry['name']))
+        return all(self.dsc.verify_file(self.workdir / entry['name'])
                    for entry in self.dsc['Files']
                    if orig_re.match(entry['name']))
 
@@ -487,7 +488,9 @@ class SourcePackage(ABC):
         if destdir:
             cmd.append(destdir)
         Logger.debug(' '.join(cmd))
-        if subprocess.call(cmd, cwd=self.workdir):
+        result = subprocess.run(cmd, cwd=str(self.workdir), encoding='utf-8',
+                                stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+        if result.returncode != 0:
             Logger.error('Source unpack failed.')
             sys.exit(1)
 
@@ -500,7 +503,7 @@ class SourcePackage(ABC):
         difffn = newpkg.dsc_name[:-3] + 'debdiff'
         Logger.debug(' '.join(cmd) + ('> %s' % difffn))
         with open(difffn, 'w') as f:
-            if subprocess.call(cmd, stdout=f, cwd=self.workdir) > 2:
+            if subprocess.call(cmd, stdout=f, cwd=str(self.workdir)) > 2:
                 Logger.error('Debdiff failed.')
                 sys.exit(1)
         if diffstat:
@@ -659,6 +662,27 @@ class PersonalPackageArchiveSourcePackage(UbuntuSourcePackage):
         Logger.debug(f"Using PPA '{ppa.web_link}'")
         return ppa
 
+    def _private_ppa_url(self, filename):
+        "Format the URL for a filename using the private PPA server"
+        url = self.getArchive().getMySubscriptionURL()
+        pkg = self.source
+        subdir = pkg[:4] if pkg.startswith('lib') else pkg[:1]
+        return f'{url}/pool/main/{subdir}/{pkg}/{filename}'
+
+    def _source_urls(self, name):
+        "Generator of sources for name"
+        if self.getArchive().private:
+            yield self._private_ppa_url(name)
+        else:
+            yield from super()._source_urls(name)
+
+    def _binary_urls(self, name, bpph):
+        "Generator of URLs for name"
+        if self.getArchive().private:
+            yield self._private_ppa_url(name)
+        else:
+            yield from super()._binary_urls(name, bpph)
+
 
 class UbuntuCloudArchiveSourcePackage(PersonalPackageArchiveSourcePackage):
     "Download / unpack an Ubuntu Cloud Archive source package"
